;;
;; SAL
;; Copyright (C) 2006, SRI International.  All Rights Reserved.
;;
;; This program is free software; you can redistribute it and/or 
;; modify it under the terms of the GNU General Public License 
;; as published by the Free Software Foundation; either version 2
;; of the License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of 
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
;; GNU General Public License for more details. 
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software 
;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
;;

;; -*- Mode: Scheme -*-

(module trace
        (include "utility.macros")
        (import xformat)
        (export (trace/enable! kind)
                (trace/disable! kind)
                (trace/store kind msg . args)
                (trace/with-output-to thunk))
        (eval (export-exports)))

(define *trace-output-port* (compile-if 
                             #f
                             (open-output-file ".salenv-trace")
                             (current-output-port)))

(define (trace/enable! kind)
  (putprop! kind 'trace-enabled #t))

(define (trace/disable! kind)
  (remprop! kind 'trace-enabled))

(define (trace/store kind msg . args)
  (unless *trace-output-port*
    (error 'trace "The program was not compiled with trace information." #unspecified))
  (apply xformat (cons* *trace-output-port* (string-append "[" (symbol->string kind) "] " msg "~%") args))
  (flush-output-port *trace-output-port*))

(define (trace/with-output-to thunk)
  (unless *trace-output-port*
    (error 'trace "The program was not compiled with trace information." #unspecified))
  (with-output-to-port *trace-output-port* thunk)
  (flush-output-port *trace-output-port*))
